From 215f385023d3700cf9177b338071b79a4f60acff Mon Sep 17 00:00:00 2001
From: Naohiro KORIYAMA <nkoriyama@gmail.com>
Date: Tue, 13 Sep 2011 19:26:21 +0900
Subject: [PATCH] TS demux: get time and length.

---
 modules/demux/ts.c |  123 +++++++++++++++++++++++++++++++++++++++++++++++++++-
 1 files changed, 122 insertions(+), 1 deletions(-)

diff --git a/modules/demux/ts.c b/modules/demux/ts.c
index 441fb9b..1bdddb6 100644
--- a/modules/demux/ts.c
+++ b/modules/demux/ts.c
@@ -336,6 +336,16 @@ struct demux_sys_t
     /* how many TS packet we read at once */
     int         i_ts_read;
 
+    /* to determine length and time */
+    mtime_t     i_first_pcr;
+    int64_t     i_first_streampos;
+    mtime_t     i_current_pcr;
+    int64_t     i_current_streampos;
+    mtime_t     i_last_pcr;
+    int64_t     i_last_streampos;
+    mtime_t     i_time;
+    int         i_mux_rate;
+
     /* All pid */
     ts_pid_t    pid[8192];
 
@@ -1196,6 +1206,10 @@ static int Control( demux_t *p_demux, int i_query, va_list args )
     {
     case DEMUX_GET_POSITION:
         pf = (double*) va_arg( args, double* );
+        if( p_sys->i_last_pcr - p_sys->i_first_pcr  > 0 )
+            *pf = (double)(p_sys->i_current_pcr - p_sys->i_first_pcr) / (double)(p_sys->i_last_pcr - p_sys->i_first_pcr);
+        else
+        {
         i64 = stream_Size( p_demux->s );
         if( i64 > 0 )
         {
@@ -1210,6 +1224,7 @@ static int Control( demux_t *p_demux, int i_query, va_list args )
             else
                 *pf = 0.0;
         }
+        }
         return VLC_SUCCESS;
     case DEMUX_SET_POSITION:
         f = (double) va_arg( args, double );
@@ -1219,8 +1234,8 @@ static int Control( demux_t *p_demux, int i_query, va_list args )
             return VLC_EGENERIC;
 
         return VLC_SUCCESS;
-#if 0
 
+#if 1
     case DEMUX_GET_TIME:
         pi64 = (int64_t*)va_arg( args, int64_t * );
         if( p_sys->i_time < 0 )
@@ -1847,6 +1862,90 @@ static void ParsePES( demux_t *p_demux, ts_pid_t *pid )
     }
 }
 
+static void GetLastPCR( demux_t* p_demux )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+    int64_t i_current_pos = stream_Tell( p_demux->s );
+
+    int64_t i_last_pos = stream_Size( p_demux->s ) - p_sys->i_packet_size;
+    int64_t i_pos = i_last_pos - p_sys->i_packet_size * 1000; /* FIXME how many packets to get last pcr? */
+    if ( i_pos < 0 )
+        return;
+
+    stream_Seek( p_demux->s, i_pos );
+
+    block_t     *p_pkt;
+    while( vlc_object_alive (p_demux) )
+    {
+        if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
+        {
+            msg_Dbg( p_demux, "eof ?" );
+            break;
+        }
+        if( p_pkt->p_buffer[0] != 0x47 )
+        {
+            msg_Warn( p_demux, "lost synchro" );
+            block_Release( p_pkt );
+            while( vlc_object_alive (p_demux) )
+            {
+                const uint8_t *p_peek;
+                int i_peek, i_skip = 0;
+
+                i_peek = stream_Peek( p_demux->s, &p_peek,
+                        p_sys->i_packet_size * 10 );
+                if( i_peek < p_sys->i_packet_size + 1 )
+                {
+                    msg_Dbg( p_demux, "eof ?" );
+                    break;
+                }
+
+                while( i_skip < i_peek - p_sys->i_packet_size )
+                {
+                    if( p_peek[i_skip] == 0x47 &&
+                            p_peek[i_skip + p_sys->i_packet_size] == 0x47 )
+                    {
+                        break;
+                    }
+                    i_skip++;
+                }
+                msg_Dbg( p_demux, "skipping %d bytes of garbage", i_skip );
+                stream_Read( p_demux->s, NULL, i_skip );
+
+                if( i_skip < i_peek - p_sys->i_packet_size )
+                {
+                    break;
+                }
+            }
+            if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
+            {
+                msg_Dbg( p_demux, "eof ?" );
+                break;
+            }
+        }
+        const uint8_t *p = p_pkt->p_buffer;
+        if( ( p[3]&0x20 ) && /* adaptation */
+                ( p[5]&0x10 ) &&
+                ( p[4] >= 7 ) )
+        {
+            /* PCR is 33 bits */
+            const mtime_t i_pcr = ( (mtime_t)p[6] << 25 ) |
+                ( (mtime_t)p[7] << 17 ) |
+                ( (mtime_t)p[8] << 9 ) |
+                ( (mtime_t)p[9] << 1 ) |
+                ( (mtime_t)p[10] >> 7 );
+
+            p_sys->i_last_pcr = i_pcr;
+            p_sys->i_last_streampos = stream_Tell( p_demux->s );
+        }
+        block_Release( p_pkt );
+
+        if( stream_Tell( p_demux->s ) >= i_last_pos ) break;
+    }
+
+    stream_Seek( p_demux->s, i_current_pos );
+    return;
+}
+
 static void PCRHandle( demux_t *p_demux, ts_pid_t *pid, block_t *p_bk )
 {
     demux_sys_t   *p_sys = p_demux->p_sys;
@@ -1866,6 +1965,28 @@ static void PCRHandle( demux_t *p_demux, ts_pid_t *pid, block_t *p_bk )
                               ( (mtime_t)p[9] << 1 ) |
                               ( (mtime_t)p[10] >> 7 );
 
+        if( p_sys->i_first_pcr == 0 )
+        {
+            p_sys->i_first_pcr = i_pcr;
+            p_sys->i_first_streampos = stream_Tell( p_demux->s );
+
+            GetLastPCR( p_demux );
+            if( p_sys->i_last_pcr - p_sys->i_first_pcr > 0 )
+            {
+                p_sys->i_mux_rate = (stream_Size( p_demux->s )) * INT64_C(1000000) / 50 / ((p_sys->i_last_pcr - p_sys->i_first_pcr) * 100 / 9);
+            }
+        }
+        else
+        {
+            p_sys->i_current_pcr = i_pcr;
+            p_sys->i_current_streampos = stream_Tell( p_demux->s );
+
+            if( p_sys->i_current_pcr > 0 && p_sys->i_first_pcr > 0 && p_sys->i_current_pcr > p_sys->i_first_pcr )
+            {
+                p_sys->i_time = (p_sys->i_current_pcr - p_sys->i_first_pcr) * 100 / 9;
+            }
+        }
+
         /* Search program and set the PCR */
         for( int i = 0; i < p_sys->i_pmt; i++ )
         {
-- 
1.7.4.4

